#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

#define NV_PERIOD 999983
#define STDDEV 0.5
#define RES 300
#define TRIALS 100000
#define MARGIN 0.25

long int nvi = 0, nvd = 1;
double nvs[NV_PERIOD];
#define normalvariate() nvs[(nvi = (nvi + nvd) % NV_PERIOD)]

void init_variates() {
    FILE* fp = fopen("random.dat", "r");
    fread(nvs, sizeof(double), NV_PERIOD, fp);
    fclose(fp);
}

long plucounts[3], appcounts[3], irvcounts[3][3], concounts[3][3];
double ax = 0, ay = 0.9999999999999;
double bx = -0.8660254037844386, by = -0.5;
double cx = 0.8660254037844386, cy = -0.5;

void clearcounts() {
    int i, j;
    for (i = 0; i < 3; i++) {
        plucounts[i] = appcounts[i] = 0;
        for (j = 0; j < 3; j++) irvcounts[i][j] = concounts[i][j] = 0;
    }
}

void countup(double x, double y, double t) {
    double da = (x - ax)*(x - ax) + (y - ay)*(y - ay);
    double db = (x - bx)*(x - bx) + (y - by)*(y - by);
    double dc = (x - cx)*(x - cx) + (y - cy)*(y - cy);
    int ab = (da <= db), ac = (da <= dc), ba = (db <= da);
    int bc = (db <= dc), ca = (dc <= da), cb = (dc <= db);

    appcounts[0] += (da <= t*t);
    appcounts[1] += (db <= t*t);
    appcounts[2] += (dc <= t*t);
}

void runsim(int trials, double x, double y, double stddev) {
    int i;
    double sx, sy, t;
    clearcounts();
    for (i = 0; i < trials; i++) {
        sx = x + normalvariate()*stddev;
        sy = y + normalvariate()*stddev;
        countup(sx, sy, exp(normalvariate())*0.25);
    }
}

#define move(s1, s2, s3, d1, d2) { \
    irvcounts[d1][d2] += irvcounts[s1][s2]; irvcounts[s1][s2] = 0; \
}

#define bitmax(a, b, c) \
    (a > b ? (a == c ? 5 : a > c ? 1 : 4) : \
     b > a ? (b == c ? 6 : b > c ? 2 : 4) : \
             (a == c ? 7 : a > c ? 3 : 4))

int appwinner() {
    int va = appcounts[0];
    int vb = appcounts[1];
    int vc = appcounts[2];
    return bitmax(va, vb, vc);
}

void makemap(double stddev, int res, int trials,
             int axx, int ayy, int bxx, int byy, int cxx, int cyy) {
    int margin = (int) (MARGIN * res);
    int i, j;
    char basename[100], filename[100];
    FILE* ffp;
    FILE* pfp;
    FILE* afp;
    FILE* ifp;
    FILE* cfp;

    ax = axx*0.01;
    ay = ayy*0.01;
    bx = bxx*0.01;
    by = byy*0.01;
    cx = cxx*0.01;
    cy = cyy*0.01;

    printf("mapping %02d,%02d %02d,%02d %02d,%02d\n",
           axx, ayy, bxx, byy, cxx, cyy);

    sprintf(basename, "%02d,%02d_%02d,%02d_%02d,%02d",
                      axx, ayy, bxx, byy, cxx, cyy);

    sprintf(filename, "%s_app.ppm", basename);
    afp = fopen(filename, "w");

    fprintf(afp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);

    for (i = res + margin - 1; i >= -margin; i--) {
        double y = i/(double) res;
        for (j = -margin; j < res + margin; j++) {
            double x = j/(double) res;
            nvi = random() % NV_PERIOD;
            nvd = (random() % (NV_PERIOD - 1)) + 1;

            runsim(trials, x, y, stddev);
            int a = appwinner();
            fputc(a & 1 ? 255 : 0, afp);
            fputc(a & 2 ? 255 : 0, afp);
            fputc(a & 4 ? 255 : 0, afp);
        }
        fflush(afp);
        fputc('.', stdout);
        fflush(stdout);
    }
    fputc('\n', stdout);
    fclose(afp);
}

#define RC (random() % 100)

int main(int argc, char* argv[]) {
    srandom(0);
    init_variates();

    if (argc == 1) {
        // makemap(STDDEV, RES, TRIALS, 50, 99,  7, 25, 93, 25);
        makemap(STDDEV, RES, TRIALS, 54, 47, 77, 64, 13, 10);
        makemap(STDDEV, RES, TRIALS, 40, 20, 89, 78, 72, 60);
        makemap(STDDEV, RES, TRIALS,  4, 17, 49,  1, 41,  2);
        makemap(STDDEV, RES, TRIALS, 37, 63, 26, 74, 83,  1);
        makemap(STDDEV, RES, TRIALS, 39, 93, 24, 58, 64,  1);
        makemap(STDDEV, RES, TRIALS, 72,  2, 30, 67, 73, 54);
        makemap(STDDEV, RES, TRIALS, 93, 49, 79, 42, 27, 45);
        makemap(STDDEV, RES, TRIALS, 98, 61, 75, 88, 19, 90);
        exit(1);
    }

    srandom(argc);
    while (1) {
        makemap(STDDEV, RES, TRIALS, RC, RC, RC, RC, RC, RC);
    }
}
